/* Copyright (c) 2016
 * The Trustees of Columbia University in the City of New York
 * All rights reserved.
 *
 * Authors:  Orestis Polychroniou  (orestis@cs.columbia.edu)
 *           Eva Sitaridi (eva@cs.columbia.edu)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <time.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>
#include <pthread.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <immintrin.h>

#include "dfa.h"


size_t regex_cnt(const uint8_t *strings,
                        size_t  tuples,
                        size_t  length,
                        size_t *loops,
                 const  int8_t *states,
                        size_t  ini_state,
                        size_t  rej_states)
{
	const uint8_t *str = strings;
	size_t i = 0, o = 0, l = 0;
	do {
		const uint8_t *str_len = &str[length];
		size_t s = ini_state;
		do {
			s <<= 8;
			s |= *str++;
			s = states[s];
			l++;
		} while (0 <= (ssize_t) s && str != str_len);
		o += ++s > rej_states ? 1 : 0;
		str = str_len;
	} while (++i != tuples);
	if (loops != NULL) *loops = l;
	return o;
}

static inline size_t __attribute__((always_inline))
inc_gt(size_t i, size_t x, size_t y)
{
	asm("cmpq	%1, %2\n\t"
	    "adcq	$0, %0"
	: "=r"(i) : "r"(x), "r"(y), "0"(i) : "cc");
	return i;
}

size_t regex_sca(const uint8_t *strings,
                      uint32_t *rids,
                        size_t  tuples,
                        size_t  length,
                 const  int8_t *states,
                        size_t  ini_state,
                        size_t  rej_states)
{
	const uint8_t *str = strings;
	size_t i = 0, o = 0;
	do {
		const uint8_t *str_len = &str[length];
		size_t s = ini_state;
		do {
			s <<= 8;
			s |= *str++;
			s = states[s];
		} while (0 <= (ssize_t) s && str != str_len);
		rids[o] = i;
		o = inc_gt(o, ++s, rej_states);
		str = str_len;
	} while (++i != tuples);
	return o;
}

#ifndef __MIC

static const uint64_t select_store[256] = { 0x0001020304050607ll,
0x0102030405060700ll, 0x0002030405060701ll, 0x0203040506070100ll,
0x0001030405060702ll, 0x0103040506070200ll, 0x0003040506070201ll,
0x0304050607020100ll, 0x0001020405060703ll, 0x0102040506070300ll,
0x0002040506070301ll, 0x0204050607030100ll, 0x0001040506070302ll,
0x0104050607030200ll, 0x0004050607030201ll, 0x0405060703020100ll,
0x0001020305060704ll, 0x0102030506070400ll, 0x0002030506070401ll,
0x0203050607040100ll, 0x0001030506070402ll, 0x0103050607040200ll,
0x0003050607040201ll, 0x0305060704020100ll, 0x0001020506070403ll,
0x0102050607040300ll, 0x0002050607040301ll, 0x0205060704030100ll,
0x0001050607040302ll, 0x0105060704030200ll, 0x0005060704030201ll,
0x0506070403020100ll, 0x0001020304060705ll, 0x0102030406070500ll,
0x0002030406070501ll, 0x0203040607050100ll, 0x0001030406070502ll,
0x0103040607050200ll, 0x0003040607050201ll, 0x0304060705020100ll,
0x0001020406070503ll, 0x0102040607050300ll, 0x0002040607050301ll,
0x0204060705030100ll, 0x0001040607050302ll, 0x0104060705030200ll,
0x0004060705030201ll, 0x0406070503020100ll, 0x0001020306070504ll,
0x0102030607050400ll, 0x0002030607050401ll, 0x0203060705040100ll,
0x0001030607050402ll, 0x0103060705040200ll, 0x0003060705040201ll,
0x0306070504020100ll, 0x0001020607050403ll, 0x0102060705040300ll,
0x0002060705040301ll, 0x0206070504030100ll, 0x0001060705040302ll,
0x0106070504030200ll, 0x0006070504030201ll, 0x0607050403020100ll,
0x0001020304050706ll, 0x0102030405070600ll, 0x0002030405070601ll,
0x0203040507060100ll, 0x0001030405070602ll, 0x0103040507060200ll,
0x0003040507060201ll, 0x0304050706020100ll, 0x0001020405070603ll,
0x0102040507060300ll, 0x0002040507060301ll, 0x0204050706030100ll,
0x0001040507060302ll, 0x0104050706030200ll, 0x0004050706030201ll,
0x0405070603020100ll, 0x0001020305070604ll, 0x0102030507060400ll,
0x0002030507060401ll, 0x0203050706040100ll, 0x0001030507060402ll,
0x0103050706040200ll, 0x0003050706040201ll, 0x0305070604020100ll,
0x0001020507060403ll, 0x0102050706040300ll, 0x0002050706040301ll,
0x0205070604030100ll, 0x0001050706040302ll, 0x0105070604030200ll,
0x0005070604030201ll, 0x0507060403020100ll, 0x0001020304070605ll,
0x0102030407060500ll, 0x0002030407060501ll, 0x0203040706050100ll,
0x0001030407060502ll, 0x0103040706050200ll, 0x0003040706050201ll,
0x0304070605020100ll, 0x0001020407060503ll, 0x0102040706050300ll,
0x0002040706050301ll, 0x0204070605030100ll, 0x0001040706050302ll,
0x0104070605030200ll, 0x0004070605030201ll, 0x0407060503020100ll,
0x0001020307060504ll, 0x0102030706050400ll, 0x0002030706050401ll,
0x0203070605040100ll, 0x0001030706050402ll, 0x0103070605040200ll,
0x0003070605040201ll, 0x0307060504020100ll, 0x0001020706050403ll,
0x0102070605040300ll, 0x0002070605040301ll, 0x0207060504030100ll,
0x0001070605040302ll, 0x0107060504030200ll, 0x0007060504030201ll,
0x0706050403020100ll, 0x0001020304050607ll, 0x0102030405060700ll,
0x0002030405060701ll, 0x0203040506070100ll, 0x0001030405060702ll,
0x0103040506070200ll, 0x0003040506070201ll, 0x0304050607020100ll,
0x0001020405060703ll, 0x0102040506070300ll, 0x0002040506070301ll,
0x0204050607030100ll, 0x0001040506070302ll, 0x0104050607030200ll,
0x0004050607030201ll, 0x0405060703020100ll, 0x0001020305060704ll,
0x0102030506070400ll, 0x0002030506070401ll, 0x0203050607040100ll,
0x0001030506070402ll, 0x0103050607040200ll, 0x0003050607040201ll,
0x0305060704020100ll, 0x0001020506070403ll, 0x0102050607040300ll,
0x0002050607040301ll, 0x0205060704030100ll, 0x0001050607040302ll,
0x0105060704030200ll, 0x0005060704030201ll, 0x0506070403020100ll,
0x0001020304060705ll, 0x0102030406070500ll, 0x0002030406070501ll,
0x0203040607050100ll, 0x0001030406070502ll, 0x0103040607050200ll,
0x0003040607050201ll, 0x0304060705020100ll, 0x0001020406070503ll,
0x0102040607050300ll, 0x0002040607050301ll, 0x0204060705030100ll,
0x0001040607050302ll, 0x0104060705030200ll, 0x0004060705030201ll,
0x0406070503020100ll, 0x0001020306070504ll, 0x0102030607050400ll,
0x0002030607050401ll, 0x0203060705040100ll, 0x0001030607050402ll,
0x0103060705040200ll, 0x0003060705040201ll, 0x0306070504020100ll,
0x0001020607050403ll, 0x0102060705040300ll, 0x0002060705040301ll,
0x0206070504030100ll, 0x0001060705040302ll, 0x0106070504030200ll,
0x0006070504030201ll, 0x0607050403020100ll, 0x0001020304050706ll,
0x0102030405070600ll, 0x0002030405070601ll, 0x0203040507060100ll,
0x0001030405070602ll, 0x0103040507060200ll, 0x0003040507060201ll,
0x0304050706020100ll, 0x0001020405070603ll, 0x0102040507060300ll,
0x0002040507060301ll, 0x0204050706030100ll, 0x0001040507060302ll,
0x0104050706030200ll, 0x0004050706030201ll, 0x0405070603020100ll,
0x0001020305070604ll, 0x0102030507060400ll, 0x0002030507060401ll,
0x0203050706040100ll, 0x0001030507060402ll, 0x0103050706040200ll,
0x0003050706040201ll, 0x0305070604020100ll, 0x0001020507060403ll,
0x0102050706040300ll, 0x0002050706040301ll, 0x0205070604030100ll,
0x0001050706040302ll, 0x0105070604030200ll, 0x0005070604030201ll,
0x0507060403020100ll, 0x0001020304070605ll, 0x0102030407060500ll,
0x0002030407060501ll, 0x0203040706050100ll, 0x0001030407060502ll,
0x0103040706050200ll, 0x0003040706050201ll, 0x0304070605020100ll,
0x0001020407060503ll, 0x0102040706050300ll, 0x0002040706050301ll,
0x0204070605030100ll, 0x0001040706050302ll, 0x0104070605030200ll,
0x0004070605030201ll, 0x0407060503020100ll, 0x0001020307060504ll,
0x0102030706050400ll, 0x0002030706050401ll, 0x0203070605040100ll,
0x0001030706050402ll, 0x0103070605040200ll, 0x0003070605040201ll,
0x0307060504020100ll, 0x0001020706050403ll, 0x0102070605040300ll,
0x0002070605040301ll, 0x0207060504030100ll, 0x0001070605040302ll,
0x0107060504030200ll, 0x0007060504030201ll, 0x0706050403020100ll};

static const uint64_t select_load[256] = {  0x0001020304050607ll,
0x0102030405060700ll, 0x0102030405060007ll, 0x0203040506070100ll,
0x0102030405000607ll, 0x0203040506010700ll, 0x0203040506010007ll,
0x0304050607020100ll, 0x0102030400050607ll, 0x0203040501060700ll,
0x0203040501060007ll, 0x0304050602070100ll, 0x0203040501000607ll,
0x0304050602010700ll, 0x0304050602010007ll, 0x0405060703020100ll,
0x0102030004050607ll, 0x0203040105060700ll, 0x0203040105060007ll,
0x0304050206070100ll, 0x0203040105000607ll, 0x0304050206010700ll,
0x0304050206010007ll, 0x0405060307020100ll, 0x0203040100050607ll,
0x0304050201060700ll, 0x0304050201060007ll, 0x0405060302070100ll,
0x0304050201000607ll, 0x0405060302010700ll, 0x0405060302010007ll,
0x0506070403020100ll, 0x0102000304050607ll, 0x0203010405060700ll,
0x0203010405060007ll, 0x0304020506070100ll, 0x0203010405000607ll,
0x0304020506010700ll, 0x0304020506010007ll, 0x0405030607020100ll,
0x0203010400050607ll, 0x0304020501060700ll, 0x0304020501060007ll,
0x0405030602070100ll, 0x0304020501000607ll, 0x0405030602010700ll,
0x0405030602010007ll, 0x0506040703020100ll, 0x0203010004050607ll,
0x0304020105060700ll, 0x0304020105060007ll, 0x0405030206070100ll,
0x0304020105000607ll, 0x0405030206010700ll, 0x0405030206010007ll,
0x0506040307020100ll, 0x0304020100050607ll, 0x0405030201060700ll,
0x0405030201060007ll, 0x0506040302070100ll, 0x0405030201000607ll,
0x0506040302010700ll, 0x0506040302010007ll, 0x0607050403020100ll,
0x0100020304050607ll, 0x0201030405060700ll, 0x0201030405060007ll,
0x0302040506070100ll, 0x0201030405000607ll, 0x0302040506010700ll,
0x0302040506010007ll, 0x0403050607020100ll, 0x0201030400050607ll,
0x0302040501060700ll, 0x0302040501060007ll, 0x0403050602070100ll,
0x0302040501000607ll, 0x0403050602010700ll, 0x0403050602010007ll,
0x0504060703020100ll, 0x0201030004050607ll, 0x0302040105060700ll,
0x0302040105060007ll, 0x0403050206070100ll, 0x0302040105000607ll,
0x0403050206010700ll, 0x0403050206010007ll, 0x0504060307020100ll,
0x0302040100050607ll, 0x0403050201060700ll, 0x0403050201060007ll,
0x0504060302070100ll, 0x0403050201000607ll, 0x0504060302010700ll,
0x0504060302010007ll, 0x0605070403020100ll, 0x0201000304050607ll,
0x0302010405060700ll, 0x0302010405060007ll, 0x0403020506070100ll,
0x0302010405000607ll, 0x0403020506010700ll, 0x0403020506010007ll,
0x0504030607020100ll, 0x0302010400050607ll, 0x0403020501060700ll,
0x0403020501060007ll, 0x0504030602070100ll, 0x0403020501000607ll,
0x0504030602010700ll, 0x0504030602010007ll, 0x0605040703020100ll,
0x0302010004050607ll, 0x0403020105060700ll, 0x0403020105060007ll,
0x0504030206070100ll, 0x0403020105000607ll, 0x0504030206010700ll,
0x0504030206010007ll, 0x0605040307020100ll, 0x0403020100050607ll,
0x0504030201060700ll, 0x0504030201060007ll, 0x0605040302070100ll,
0x0504030201000607ll, 0x0605040302010700ll, 0x0605040302010007ll,
0x0706050403020100ll, 0x0001020304050607ll, 0x0102030405060700ll,
0x0102030405060007ll, 0x0203040506070100ll, 0x0102030405000607ll,
0x0203040506010700ll, 0x0203040506010007ll, 0x0304050607020100ll,
0x0102030400050607ll, 0x0203040501060700ll, 0x0203040501060007ll,
0x0304050602070100ll, 0x0203040501000607ll, 0x0304050602010700ll,
0x0304050602010007ll, 0x0405060703020100ll, 0x0102030004050607ll,
0x0203040105060700ll, 0x0203040105060007ll, 0x0304050206070100ll,
0x0203040105000607ll, 0x0304050206010700ll, 0x0304050206010007ll,
0x0405060307020100ll, 0x0203040100050607ll, 0x0304050201060700ll,
0x0304050201060007ll, 0x0405060302070100ll, 0x0304050201000607ll,
0x0405060302010700ll, 0x0405060302010007ll, 0x0506070403020100ll,
0x0102000304050607ll, 0x0203010405060700ll, 0x0203010405060007ll,
0x0304020506070100ll, 0x0203010405000607ll, 0x0304020506010700ll,
0x0304020506010007ll, 0x0405030607020100ll, 0x0203010400050607ll,
0x0304020501060700ll, 0x0304020501060007ll, 0x0405030602070100ll,
0x0304020501000607ll, 0x0405030602010700ll, 0x0405030602010007ll,
0x0506040703020100ll, 0x0203010004050607ll, 0x0304020105060700ll,
0x0304020105060007ll, 0x0405030206070100ll, 0x0304020105000607ll,
0x0405030206010700ll, 0x0405030206010007ll, 0x0506040307020100ll,
0x0304020100050607ll, 0x0405030201060700ll, 0x0405030201060007ll,
0x0506040302070100ll, 0x0405030201000607ll, 0x0506040302010700ll,
0x0506040302010007ll, 0x0607050403020100ll, 0x0100020304050607ll,
0x0201030405060700ll, 0x0201030405060007ll, 0x0302040506070100ll,
0x0201030405000607ll, 0x0302040506010700ll, 0x0302040506010007ll,
0x0403050607020100ll, 0x0201030400050607ll, 0x0302040501060700ll,
0x0302040501060007ll, 0x0403050602070100ll, 0x0302040501000607ll,
0x0403050602010700ll, 0x0403050602010007ll, 0x0504060703020100ll,
0x0201030004050607ll, 0x0302040105060700ll, 0x0302040105060007ll,
0x0403050206070100ll, 0x0302040105000607ll, 0x0403050206010700ll,
0x0403050206010007ll, 0x0504060307020100ll, 0x0302040100050607ll,
0x0403050201060700ll, 0x0403050201060007ll, 0x0504060302070100ll,
0x0403050201000607ll, 0x0504060302010700ll, 0x0504060302010007ll,
0x0605070403020100ll, 0x0201000304050607ll, 0x0302010405060700ll,
0x0302010405060007ll, 0x0403020506070100ll, 0x0302010405000607ll,
0x0403020506010700ll, 0x0403020506010007ll, 0x0504030607020100ll,
0x0302010400050607ll, 0x0403020501060700ll, 0x0403020501060007ll,
0x0504030602070100ll, 0x0403020501000607ll, 0x0504030602010700ll,
0x0504030602010007ll, 0x0605040703020100ll, 0x0302010004050607ll,
0x0403020105060700ll, 0x0403020105060007ll, 0x0504030206070100ll,
0x0403020105000607ll, 0x0504030206010700ll, 0x0504030206010007ll,
0x0605040307020100ll, 0x0403020100050607ll, 0x0504030201060700ll,
0x0504030201060007ll, 0x0605040302070100ll, 0x0504030201000607ll,
0x0605040302010700ll, 0x0605040302010007ll, 0x0706050403020100ll};

static inline size_t __attribute__((always_inline))
_mm256_movemask_epi32(__m256i x)
{
	size_t r;
	asm("vmovmskps	%1, %0" : "=r"(r) : "x"(x));
	return r;
}

size_t regex_vec_1(const uint8_t *strings,
                        uint32_t *rids,
                          size_t  tuples,
                          size_t  length,
                   const  int8_t *states,
                          size_t  ini_state,
                          size_t  rej_states)
{
	const long long *strings_long = (const long long*) strings;
	const int32_t    *states_int  = (const int32_t*) states;
	// constant masks
	const __m256i m0 = _mm256_setzero_si256();
	const __m256i mFF = _mm256_set1_epi32(255);
	const __m256i ord = _mm256_set_epi32(7,5,3,1,6,4,2,0);
	const __m256i len = _mm256_set1_epi32(length);
	const __m256i ini = _mm256_set1_epi32(ini_state);
	const __m256i rej = _mm256_set1_epi32(rej_states);
	// maintain out-of-order state
	const size_t loops_1 = length <= 4 ? length : 4;
	const size_t loops_2 = length <= 4 ? 0 :
	                       length <= 8 ? length - 4 : 4;
	size_t i = 8, o = 0;
	__m256i rid = _mm256_set_epi32(7,6,5,4,3,2,1,0);
	__m256i cur = ini, rem = len;
	if (tuples > 8) for (;;) {
		// compute (byte) offset using rid and string offset
		__m256i ptr = _mm256_mullo_epi32(rid, len);
		ptr = _mm256_add_epi32(ptr, _mm256_sub_epi32(len, rem));
		// gather 8 bytes per string
		__m256i word_L = _mm256_i32gather_epi64(strings_long, _mm256_castsi256_si128(ptr), 1);
		ptr = _mm256_permute4x64_epi64(ptr, _MM_SHUFFLE(1,0,3,2));
		__m256i word_H = _mm256_i32gather_epi64(strings_long, _mm256_castsi256_si128(ptr), 1);
		word_L = _mm256_permutevar8x32_epi32(word_L, ord);
		word_H = _mm256_permutevar8x32_epi32(word_H, ord);
		__m256i word_1 = _mm256_permute2f128_si256(word_L, word_H, _MM_SHUFFLE(0,2,0,0));
		__m256i word_2 = _mm256_permute2f128_si256(word_L, word_H, _MM_SHUFFLE(0,3,0,1));
		// process first 4 bytes
		size_t l = loops_1;
		do {
			// find valid lanes
			__m256i val = _mm256_cmpgt_epi32(rem, m0);
#ifdef _NO_LANE_VALID
			if (_mm256_testz_si256(val, val)) goto no_lane_valid;
#endif
			// compute state transition index
			__m256i nex = _mm256_slli_epi32(cur, 8);
			nex = _mm256_or_si256(nex, _mm256_and_si256(word_1, mFF));
			// gather new state for valid lanes
			cur = _mm256_mask_i32gather_epi32(cur, states_int, nex, val, 1);
			// clear processed byte and update remaining bytes
			rem = _mm256_add_epi32(rem, val);
			word_1 = _mm256_srli_epi32(word_1, 8);
			// clear high order bits of state
			cur = _mm256_slli_epi32(cur, 24);
			cur = _mm256_srai_epi32(cur, 24);
			// find which strings are done and reset remaining bytes
			rem = _mm256_andnot_si256(_mm256_cmpgt_epi32(m0, cur), rem);
		} while (--l);
		if ((l = loops_2)) do {
			// find valid lanes
			__m256i val = _mm256_cmpgt_epi32(rem, m0);
#ifdef _NO_LANE_VALID
			if (_mm256_testz_si256(val, val)) goto no_lane_valid;
#endif
			// find transition offset
			__m256i nex = _mm256_slli_epi32(cur, 8);
			nex = _mm256_or_si256(nex, _mm256_and_si256(word_2, mFF));
			// gather new state for valid lanes
			cur = _mm256_mask_i32gather_epi32(cur, states_int, nex, val, 1);
			// clear processed byte and update remaining bytes
			rem = _mm256_add_epi32(rem, val);
			word_2 = _mm256_srli_epi32(word_2, 8);
			// clear high order bits of state
			cur = _mm256_slli_epi32(cur, 24);
			cur = _mm256_srai_epi32(cur, 24);
			// find which strings are done and reset remaining bytes
			rem = _mm256_andnot_si256(_mm256_cmpgt_epi32(m0, cur), rem);
		} while (--l);
#ifdef _NO_LANE_VALID
no_lane_valid:;
#endif
		// find processed and accepted strings
		__m256i inv = _mm256_cmpeq_epi32(rem, m0);
		__m256i acc = _mm256_cmpgt_epi32(rej, cur);
		acc = _mm256_andnot_si256(acc, inv);
		acc = _mm256_or_si256(acc, _mm256_cmpgt_epi32(inv, cur));
		size_t mi = _mm256_movemask_epi32(inv);
		size_t ma = _mm256_movemask_epi32(acc);
		// store rids of accepted strings
		if (ma) {
			__m128i *t = (__m128i*) &select_store[ma];
			acc = _mm256_cvtepi8_epi32(_mm_loadl_epi64(t));
			acc = _mm256_permutevar8x32_epi32(rid, acc);
			_mm256_storeu_si256((__m256i*) &rids[o], acc);
			o += _mm_popcnt_u64(ma);
		}
		// store rids of processed strings
		if (mi) {
			__m128i *t = (__m128i*) &select_load[mi];
			__m256i per = _mm256_cvtepi8_epi32(_mm_loadl_epi64(t));
			__m256i nex = _mm256_broadcastd_epi32(_mm_cvtsi64_si128(i));
			__m256i rem_I = _mm256_and_si256(inv, len);
			__m256i cur_I = _mm256_and_si256(inv, ini);
			nex = _mm256_add_epi32(nex, per);
			cur = _mm256_andnot_si256(inv, cur);
			rid = _mm256_andnot_si256(inv, rid);
			nex = _mm256_and_si256(inv, nex);
			rem = _mm256_or_si256(rem, rem_I);
			cur = _mm256_or_si256(cur, cur_I);
			rid = _mm256_or_si256(rid, nex);
			i += _mm_popcnt_u64(mi);
			if (i >= tuples) break;
		}
	}
	// process last strings with scalar code
	uint32_t buf[8];
	_mm256_storeu_si256((__m256i*) buf, rid);
	for (i = 0 ; i != 8 ; ++i) {
		size_t j = buf[i];
		if (j >= tuples) continue;
		const uint8_t *str = &strings[j * length];
		const uint8_t *str_len = &str[length];
		size_t s = ini_state;
		do {
			s <<= 8;
			s |= *str++;
			s = states[s];
		} while (0 <= (ssize_t) s && str != str_len);
		if (++s > rej_states) rids[o++] = j;
	}
	return o;
}

size_t regex_vec_2(const uint8_t *strings,
                        uint32_t *rids,
                          size_t  tuples,
                          size_t  length,
                   const  int8_t *states,
                          size_t  ini_state,
                          size_t  rej_states)
{
	assert(tuples > 32);
	const long long *strings_long = (const long long*) strings;
	const int32_t    *states_int  = (const int32_t*) states;
	// constant masks
	const __m256i m0 = _mm256_setzero_si256();
	const __m256i m3 = _mm256_set1_epi32(3);
	const __m256i mFF = _mm256_set1_epi32(255);
	const __m256i ord = _mm256_set_epi32(7,5,3,1,6,4,2,0);
	const __m256i len = _mm256_set1_epi32(length);
	const __m256i ini = _mm256_set1_epi32(ini_state);
	const __m256i rej = _mm256_set1_epi32(rej_states);
	// maintain out-of-order state
	const size_t loops_1 = length <= 4 ? length : 4;
	const size_t loops_2 = length <= 4 ? 0 :
	                       length <= 8 ? length - 4 : 4;
	size_t i = 8, o = 0, j = tuples - 9;
	__m256i rid_L = _mm256_set_epi32(7,6,5,4,3,2,1,0);
	__m256i rid_H = _mm256_add_epi32(rid_L, _mm256_set1_epi32(j));
	__m256i cur_L = ini, cur_H = ini, rem_L = len, rem_H = len;
	for (;;) {
		// compute (byte) offset using rid and string offset
		__m256i ptr_L = _mm256_mullo_epi32(rid_L, len);
		__m256i ptr_H = _mm256_mullo_epi32(rid_H, len);
		ptr_L = _mm256_add_epi32(ptr_L, _mm256_sub_epi32(len, rem_L));
		ptr_H = _mm256_add_epi32(ptr_H, _mm256_sub_epi32(len, rem_H));
		// gather 8 bytes per string
		__m256i word_L_L = _mm256_i32gather_epi64(strings_long, _mm256_castsi256_si128(ptr_L), 1);
		__m256i word_H_L = _mm256_i32gather_epi64(strings_long, _mm256_castsi256_si128(ptr_H), 1);
		ptr_L = _mm256_permute4x64_epi64(ptr_L, _MM_SHUFFLE(1,0,3,2));
		ptr_H = _mm256_permute4x64_epi64(ptr_H, _MM_SHUFFLE(1,0,3,2));
		__m256i word_L_H = _mm256_i32gather_epi64(strings_long, _mm256_castsi256_si128(ptr_L), 1);
		__m256i word_H_H = _mm256_i32gather_epi64(strings_long, _mm256_castsi256_si128(ptr_H), 1);
		word_L_L = _mm256_permutevar8x32_epi32(word_L_L, ord);
		word_H_L = _mm256_permutevar8x32_epi32(word_H_L, ord);
		word_L_H = _mm256_permutevar8x32_epi32(word_L_H, ord);
		word_H_H = _mm256_permutevar8x32_epi32(word_H_H, ord);
		__m256i word_1_L = _mm256_permute2f128_si256(word_L_L, word_L_H, _MM_SHUFFLE(0,2,0,0));
		__m256i word_2_L = _mm256_permute2f128_si256(word_L_L, word_L_H, _MM_SHUFFLE(0,3,0,1));
		__m256i word_1_H = _mm256_permute2f128_si256(word_H_L, word_H_H, _MM_SHUFFLE(0,2,0,0));
		__m256i word_2_H = _mm256_permute2f128_si256(word_H_L, word_H_H, _MM_SHUFFLE(0,3,0,1));
		// process first 4 bytes
		size_t l = loops_1;
		do {
			// find valid lanes
			__m256i val_L = _mm256_cmpgt_epi32(rem_L, m0);
			__m256i val_H = _mm256_cmpgt_epi32(rem_H, m0);
#ifdef _NO_LANE_VALID
			__m256i val = _mm256_or_si256(val_L, val_H);
			if (_mm256_testz_si256(val, val)) goto no_lane_valid;
#endif
			// find transition offset
			__m256i nex_L = _mm256_slli_epi32(cur_L, 8);
			__m256i nex_H = _mm256_slli_epi32(cur_H, 8);
			nex_L = _mm256_or_si256(nex_L, _mm256_and_si256(word_1_L, mFF));
			nex_H = _mm256_or_si256(nex_H, _mm256_and_si256(word_1_H, mFF));
			// gather new state for valid lanes
			cur_L = _mm256_mask_i32gather_epi32(cur_L, states_int, nex_L, val_L, 1);
			cur_H = _mm256_mask_i32gather_epi32(cur_H, states_int, nex_H, val_H, 1);
			// shift used to cast char to int
			__m256i shl_L = _mm256_and_si256(nex_L, val_L);
			__m256i shl_H = _mm256_and_si256(nex_H, val_H);
			shl_L = _mm256_andnot_si256(shl_L, m3);
			shl_H = _mm256_andnot_si256(shl_H, m3);
			shl_L = _mm256_slli_epi32(shl_L, 3);
			shl_H = _mm256_slli_epi32(shl_H, 3);
			// clear processed byte and update remaining bytes
			rem_L = _mm256_add_epi32(rem_L, val_L);
			rem_H = _mm256_add_epi32(rem_H, val_H);
			word_1_L = _mm256_srli_epi32(word_1_L, 8);
			word_1_H = _mm256_srli_epi32(word_1_H, 8);
			// clear high order bits
			cur_L = _mm256_slli_epi32(cur_L, 24);
			cur_H = _mm256_slli_epi32(cur_H, 24);
			cur_L = _mm256_srai_epi32(cur_L, 24);
			cur_H = _mm256_srai_epi32(cur_H, 24);
			// find which strings are done and reset remaining bytes
			rem_L = _mm256_andnot_si256(_mm256_cmpgt_epi32(m0, cur_L), rem_L);
			rem_H = _mm256_andnot_si256(_mm256_cmpgt_epi32(m0, cur_H), rem_H);
		} while (--l);
		if ((l = loops_2)) do {
			// find valid lanes
			__m256i val_L = _mm256_cmpgt_epi32(rem_L, m0);
			__m256i val_H = _mm256_cmpgt_epi32(rem_H, m0);
#ifdef _NO_LANE_VALID
			__m256i val = _mm256_or_si256(val_L, val_H);
			if (_mm256_testz_si256(val, val)) goto no_lane_valid;
#endif
			// find transition offset
			__m256i nex_L = _mm256_slli_epi32(cur_L, 8);
			__m256i nex_H = _mm256_slli_epi32(cur_H, 8);
			nex_L = _mm256_or_si256(nex_L, _mm256_and_si256(word_2_L, mFF));
			nex_H = _mm256_or_si256(nex_H, _mm256_and_si256(word_2_H, mFF));
			// gather new state for valid lanes
			cur_L = _mm256_mask_i32gather_epi32(cur_L, states_int, nex_L, val_L, 1);
			cur_H = _mm256_mask_i32gather_epi32(cur_H, states_int, nex_H, val_H, 1);
			// shift used to cast char to int
			__m256i shl_L = _mm256_and_si256(nex_L, val_L);
			__m256i shl_H = _mm256_and_si256(nex_H, val_H);
			shl_L = _mm256_andnot_si256(shl_L, m3);
			shl_H = _mm256_andnot_si256(shl_H, m3);
			shl_L = _mm256_slli_epi32(shl_L, 3);
			shl_H = _mm256_slli_epi32(shl_H, 3);
			// clear processed byte and update remaining bytes
			rem_L = _mm256_add_epi32(rem_L, val_L);
			rem_H = _mm256_add_epi32(rem_H, val_H);
			word_2_L = _mm256_srli_epi32(word_2_L, 8);
			word_2_H = _mm256_srli_epi32(word_2_H, 8);
			// clear high order bits
			cur_L = _mm256_slli_epi32(cur_L, 24);
			cur_H = _mm256_slli_epi32(cur_H, 24);
			cur_L = _mm256_srai_epi32(cur_L, 24);
			cur_H = _mm256_srai_epi32(cur_H, 24);
			// find which strings are done and reset remaining bytes
			rem_L = _mm256_andnot_si256(_mm256_cmpgt_epi32(m0, cur_L), rem_L);
			rem_H = _mm256_andnot_si256(_mm256_cmpgt_epi32(m0, cur_H), rem_H);
		} while (--l);
#ifdef _NO_LANE_VALID
no_lane_valid:;
#endif
		// find processed and accepted strings
		__m256i inv_L = _mm256_cmpeq_epi32(rem_L, m0);
		__m256i inv_H = _mm256_cmpeq_epi32(rem_H, m0);
		__m256i acc_L = _mm256_cmpgt_epi32(rej, cur_L);
		__m256i acc_H = _mm256_cmpgt_epi32(rej, cur_H);
		acc_L = _mm256_andnot_si256(acc_L, inv_L);
		acc_H = _mm256_andnot_si256(acc_H, inv_H);
		acc_L = _mm256_or_si256(acc_L, _mm256_cmpgt_epi32(inv_L, cur_L));
		acc_H = _mm256_or_si256(acc_H, _mm256_cmpgt_epi32(inv_H, cur_H));
		size_t ma_L = _mm256_movemask_epi32(acc_L);
		size_t ma_H = _mm256_movemask_epi32(acc_H);
		if (ma_L | ma_H) {
			// load permutation mask
			__m128i *t_L = (__m128i*) &select_store[ma_L];
			__m128i *t_H = (__m128i*) &select_store[ma_H];
			acc_L = _mm256_cvtepi8_epi32(_mm_loadl_epi64(t_L));
			acc_H = _mm256_cvtepi8_epi32(_mm_loadl_epi64(t_H));
			// permute and store accepted string rids
			acc_L = _mm256_permutevar8x32_epi32(rid_L, acc_L);
			acc_H = _mm256_permutevar8x32_epi32(rid_H, acc_H);
			_mm256_storeu_si256((__m256i*) &rids[o], acc_L);
			o += _mm_popcnt_u64(ma_L);
			_mm256_storeu_si256((__m256i*) &rids[o], acc_H);
			o += _mm_popcnt_u64(ma_H);
		}
		size_t mi_L = _mm256_movemask_epi32(inv_L);
		size_t mi_H = _mm256_movemask_epi32(inv_H);
		if (mi_L | mi_H) {
			j -= _mm_popcnt_u64(mi_H);
			__m128i *t_L = (__m128i*) &select_load[mi_L];
			__m128i *t_H = (__m128i*) &select_load[mi_H];
			__m256i per_L = _mm256_cvtepi8_epi32(_mm_loadl_epi64(t_L));
			__m256i per_H = _mm256_cvtepi8_epi32(_mm_loadl_epi64(t_H));
			__m256i nex_L = _mm256_broadcastd_epi32(_mm_cvtsi64_si128(i));
			__m256i nex_H = _mm256_broadcastd_epi32(_mm_cvtsi64_si128(j));
			i += _mm_popcnt_u64(mi_L);
			__m256i len_L = _mm256_and_si256(inv_L, len);
			__m256i len_H = _mm256_and_si256(inv_H, len);
			__m256i ini_L = _mm256_and_si256(inv_L, ini);
			__m256i ini_H = _mm256_and_si256(inv_H, ini);
			nex_L = _mm256_add_epi32(nex_L, per_L);
			nex_H = _mm256_add_epi32(nex_H, per_H);
			cur_L = _mm256_andnot_si256(inv_L, cur_L);
			cur_H = _mm256_andnot_si256(inv_H, cur_H);
			rid_L = _mm256_andnot_si256(inv_L, rid_L);
			rid_H = _mm256_andnot_si256(inv_H, rid_H);
			nex_L = _mm256_and_si256(inv_L, nex_L);
			nex_H = _mm256_and_si256(inv_H, nex_H);
			rem_L = _mm256_or_si256(rem_L, len_L);
			rem_H = _mm256_or_si256(rem_H, len_H);
			cur_L = _mm256_or_si256(cur_L, ini_L);
			cur_H = _mm256_or_si256(cur_H, ini_H);
			rid_L = _mm256_or_si256(rid_L, nex_L);
			rid_H = _mm256_or_si256(rid_H, nex_H);
			if (j < i + 16) break;
		}
	}
	// store last rids in buffer
	uint32_t buf[32];
	assert(i <= j);
	_mm256_storeu_si256((__m256i*) &buf[0], rid_L);
	_mm256_storeu_si256((__m256i*) &buf[8], rid_H);
	buf[16] = tuples - 1;
	size_t l = 17;
	while (i != j) buf[l++] = i++;
	assert(l <= 32);
	// process last rids in scalar code
	for (i = 0 ; i != l ; ++i) {
		j = buf[i];
		const uint8_t *str = &strings[j * length];
		const uint8_t *str_len = &str[length];
		size_t s = ini_state;
		do {
			s <<= 8;
			s |= *str++;
			s = states[s];
		} while (0 <= (ssize_t) s && str != str_len);
		if (++s > rej_states) rids[o++] = j;
	}
	return o;
}

#ifndef _INTEL_COMPILER

static inline __m256i __attribute__((always_inline))
_mm256_loadu2_m128i(__m128i const* hiaddr, __m128i const* loaddr)
{
	__m256i lo = _mm256_castsi128_si256(_mm_loadu_si128(loaddr));
	__m256i hi = _mm256_castsi128_si256(_mm_loadu_si128(hiaddr));
	return _mm256_permute2x128_si256(lo, hi, _MM_SHUFFLE(0,2,0,0));
}

#endif

static inline size_t __attribute__((always_inline))
min(size_t x, size_t y)
{
	asm("cmpq	%0, %1\n\t"
	    "cmovcq	%1, %0"
	: "=r"(x) : "r"(y), "0"(x) : "cc");
	return x;
}

size_t regex_vec_3(const uint8_t *strings,
                        uint32_t *rids,
                          size_t  tuples,
                          size_t  length,
                   const  int8_t *states,
                          size_t  ini_state,
                          size_t  rej_states)
{
	assert((tuples & 15) == 0);
	const int32_t    *states_int  = (const int32_t*) states;
	const __m256i m16 = _mm256_set1_epi32(16);
	const __m256i m1s = _mm256_set1_epi32(-1);
	const __m256i mFF = _mm256_set1_epi32(255);
	const __m256i ini = _mm256_set1_epi32(ini_state);
	const __m256i rej = _mm256_set1_epi32(rej_states);
	__m256i r_L = _mm256_set_epi32( 7, 6, 5, 4, 3, 2, 1, 0);
	__m256i r_H = _mm256_set_epi32(15,14,13,12,11,10, 9, 8);
	uint32_t *rids_orig = rids;
	if (tuples) do {
		const uint8_t *strings_16_len = &strings[length << 4];
		__m256i s_L = ini, s_H = ini;
		size_t l = 0;
		do {
			const uint8_t *d_L =  strings;
			const uint8_t *d_H = &strings[length << 2];
			__m256i d1 = _mm256_loadu2_m128i((__m128i*) d_H, (__m128i*) d_L);
			d_L += length, d_H += length;
			__m256i d2 = _mm256_loadu2_m128i((__m128i*) d_H, (__m128i*) d_L);
			d_L += length, d_H += length;
			__m256i d3 = _mm256_loadu2_m128i((__m128i*) d_H, (__m128i*) d_L);
			d_L += length, d_H += length;
			__m256i d4 = _mm256_loadu2_m128i((__m128i*) d_H, (__m128i*) d_L);
			d_L = &strings[length << 3];
			d_H = &d_L[length << 2];
			__m256i d5 = _mm256_loadu2_m128i((__m128i*) d_H, (__m128i*) d_L);
			d_L += length, d_H += length;
			__m256i d6 = _mm256_loadu2_m128i((__m128i*) d_H, (__m128i*) d_L);
			d_L += length, d_H += length;
			__m256i d7 = _mm256_loadu2_m128i((__m128i*) d_H, (__m128i*) d_L);
			d_L += length, d_H += length;
			__m256i d8 = _mm256_loadu2_m128i((__m128i*) d_H, (__m128i*) d_L);
			size_t dl = min(length - l, 16);
			for (;;) {
				// pack characters
				__m256i d12 = _mm256_unpacklo_epi32(d1, d2);
				__m256i d34 = _mm256_unpacklo_epi32(d3, d4);
				__m256i d56 = _mm256_unpacklo_epi32(d5, d6);
				__m256i d78 = _mm256_unpacklo_epi32(d7, d8);
				__m256i d_L = _mm256_unpacklo_epi64(d12, d34);
				__m256i d_H = _mm256_unpacklo_epi64(d56, d78);
				// compute the transition offset
				__m256i n_L = _mm256_slli_epi32(s_L, 8);
				__m256i n_H = _mm256_slli_epi32(s_H, 8);
				d_L = _mm256_and_si256(d_L, mFF);
				d_H = _mm256_and_si256(d_H, mFF);
				n_L = _mm256_or_si256(n_L, d_L);
				n_H = _mm256_or_si256(n_H, d_H);
				// find which lanes are valid
				__m256i v_L = _mm256_cmpgt_epi32(s_L, m1s);
				__m256i v_H = _mm256_cmpgt_epi32(s_H, m1s);
				s_L = _mm256_mask_i32gather_epi32(s_L, states_int, n_L, v_L, 1);
				s_H = _mm256_mask_i32gather_epi32(s_H, states_int, n_H, v_H, 1);
				s_L = _mm256_slli_epi32(s_L, 24);
				s_H = _mm256_slli_epi32(s_H, 24);
				s_L = _mm256_srai_epi32(s_L, 24);
				s_H = _mm256_srai_epi32(s_H, 24);
				if (--dl == 0) break;
				// shift strings
				d1 = _mm256_srli_si256(d1, 1);
				d2 = _mm256_srli_si256(d2, 1);
				d3 = _mm256_srli_si256(d3, 1);
				d4 = _mm256_srli_si256(d4, 1);
				d5 = _mm256_srli_si256(d5, 1);
				d6 = _mm256_srli_si256(d6, 1);
				d7 = _mm256_srli_si256(d7, 1);
				d8 = _mm256_srli_si256(d8, 1);
			}
			strings += 16;
		} while ((l = l + 16) < length);
		// find accepted strings
		s_L = _mm256_sub_epi32(s_L, m1s);
		s_H = _mm256_sub_epi32(s_H, m1s);
		__m256i a_L = _mm256_cmpeq_epi32(s_L, m1s);
		__m256i a_H = _mm256_cmpeq_epi32(s_H, m1s);
		s_L = _mm256_cmpgt_epi32(s_L, rej);
		s_H = _mm256_cmpgt_epi32(s_H, rej);
		s_L = _mm256_or_si256(s_L, a_L);
		s_H = _mm256_or_si256(s_H, a_H);
		size_t m_L = _mm256_movemask_epi32(s_L);
		size_t m_H = _mm256_movemask_epi32(s_H);
		if (m_L | m_H) {
			// load permutation mask
			__m128i *t_L = (__m128i*) &select_store[m_L];
			__m128i *t_H = (__m128i*) &select_store[m_H];
			s_L = _mm256_cvtepi8_epi32(_mm_loadl_epi64(t_L));
			s_H = _mm256_cvtepi8_epi32(_mm_loadl_epi64(t_H));
			// permute and store accepted string rids
			s_L = _mm256_permutevar8x32_epi32(r_L, s_L);
			s_H = _mm256_permutevar8x32_epi32(r_H, s_H);
			_mm256_storeu_si256((__m256i*) rids, s_L);
			rids += _mm_popcnt_u64(m_L);
			_mm256_storeu_si256((__m256i*) rids, s_H);
			rids += _mm_popcnt_u64(m_H);
		}
		// increment the pointers
		r_L = _mm256_add_epi32(r_L, m16);
		r_H = _mm256_add_epi32(r_H, m16);
		strings = strings_16_len;
		tuples -= 16;
	} while (tuples);
	return rids - rids_orig;
}

#else

static inline __m512i __attribute__((always_inline))
_mm512_broadcast_epi32(const int32_t *p)
{
	return _mm512_extload_epi32(p, _MM_UPCONV_EPI32_NONE,
	                               _MM_BROADCAST_1X16, 0);
}

size_t regex_vec_1(const uint8_t *strings,
                        uint32_t *rids,
                          size_t  tuples,
                          size_t  length,
                   const  int8_t *states,
                          size_t  ini_state,
                          size_t  rej_states)
{
	// constant vector masks
	const __mmask16 k0 = _mm512_int2mask(0);
	const __m512i m1s = _mm512_set1_epi32(-1);
	const __m512i m32 = _mm512_set1_epi32(32);
	const __m512i mFF = _mm512_set1_epi32(255);
	const __m512i len = _mm512_set1_epi32(length);
	const __m512i ini = _mm512_set1_epi32(ini_state);
	const __m512i rej = _mm512_set1_epi32(rej_states);
	const __m512i inc = _mm512_set_epi32(15,14,13,12,11,10, 9, 8,
	                                      7, 6, 5, 4, 3, 2, 1, 0);
	// aligned buffer
	int32_t buf_unaligned[31];
	int32_t *buf = (int32_t*) ((63 + (size_t) buf_unaligned) & -64);
	// maintain out-of-order state
	__m512i rid = inc, cur = ini, off = _mm512_set1_epi32(0);
	const size_t loops_1 = length <= 4 ? length : 4;
	const size_t loops_2 = length <= 4 ? 0 :
	                      (length & 1) ? 1 :
	                      (length & 3) ? 2 : 4;
	size_t i = 16, o = 0;
	if (tuples > 16) for (;;) {
		// compute (byte) offset using rid and string offset
		__m512i ptr = _mm512_fmadd_epi32(rid, len, off);
		// gather 8 bytes per string
		__m512i ind = _mm512_srli_epi32(ptr, 2);
		__m512i word_1 = _mm512_i32gather_epi32(ind, &strings[0], 4);
		__m512i word_2 = _mm512_i32gather_epi32(ind, &strings[4], 4);
		if (length & 3) {
			// compute shifts
			__m512i shr = _mm512_slli_epi32(ptr, 30);
			shr = _mm512_srli_epi32(shr, 27);
			__m512i shl = _mm512_sub_epi32(m32, shr);
			// combine into words
			word_1 = _mm512_srlv_epi32(word_1, shr);
			__m512i word_3 = _mm512_sllv_epi32(word_2, shl);
			word_2 = _mm512_srlv_epi32(word_2, shr);
			word_1 = _mm512_or_epi32(word_1, word_3);
		}
		// process 5 or 6 bytes per string
		size_t l = loops_1;
		__mmask16 k = _mm512_kxnor(k0, k0);
		do {
			// compute state transition index
			__m512i nex = _mm512_slli_epi32(cur, 8);
			nex = _mm512_or_epi32(nex, _mm512_and_epi32(word_1, mFF));
			// gather new states for valid lanes
			cur = _mm512_mask_i32extgather_epi32(cur, k, nex, states,
			      _MM_UPCONV_EPI32_SINT8, 1, 0);
			// clear processed byte and update remaining bytes
			off = _mm512_mask_sub_epi32(off, k, off, m1s);
			word_1 = _mm512_srli_epi32(word_1, 8);
			// update valid lanes
			k = _mm512_mask_cmpgt_epi32_mask(k, len, off);
			k = _mm512_mask_cmpgt_epi32_mask(k, cur, m1s);
#ifdef _NO_LANE_VALID
			if (_mm512_kortestz(k, k)) goto no_lane_valid;
#endif
		} while (--l);
		if ((l = loops_2)) do {
			// compute state transition index
			__m512i nex = _mm512_slli_epi32(cur, 8);
			nex = _mm512_or_epi32(nex, _mm512_and_epi32(word_2, mFF));
			// gather new state for valid lanes
			cur = _mm512_mask_i32extgather_epi32(cur, k, nex, states,
			      _MM_UPCONV_EPI32_SINT8, 1, 0);
			// clear processed byte and update remaining bytes
			off = _mm512_mask_sub_epi32(off, k, off, m1s);
			word_2 = _mm512_srli_epi32(word_2, 8);
			// update valid lanes
			k = _mm512_mask_cmpgt_epi32_mask(k, len, off);
			k = _mm512_mask_cmpgt_epi32_mask(k, cur, m1s);
#ifdef _NO_LANE_VALID
			if (_mm512_kortestz(k, k)) goto no_lane_valid;
#endif
		} while (--l);
#ifdef _NO_LANE_VALID
no_lane_valid:;
#endif
		// store rids of accepted strings
		__m512i out = _mm512_sub_epi32(cur, m1s);
		__mmask16 ki = _mm512_knot(k);
		__mmask16 ko = _mm512_mask_cmpgt_epu32_mask(ki, out, rej);
		// store rids of accepted strings
		if (!_mm512_kortestz(ko, ko)) {
			_mm512_mask_packstorelo_epi32(&rids[o +  0], ko, rid);
			_mm512_mask_packstorehi_epi32(&rids[o + 16], ko, rid);
			o += _mm_countbits_64(_mm512_kconcatlo_64(ko, k0));
		}
		// replace rids of processed strings
		if (!_mm512_kortestz(ki, ki)) {
			// update buffer with new rids
			buf[0] = i;
			__m512i nex = _mm512_broadcast_epi32(buf);
			nex = _mm512_add_epi32(nex, inc);
			_mm512_store_epi32(buf, nex);
			i += _mm_countbits_64(_mm512_kconcatlo_64(ki, k0));
			// replace finished rids
			rid = _mm512_mask_loadunpacklo_epi32(rid, ki, buf);
			off = _mm512_mask_xor_epi32(off, ki, off, off);
			cur = _mm512_mask_blend_epi32(ki, cur, ini);
			// stop if end of tuples is reached
			if (i >= tuples) break;
		}
	}
	// process last strings with scalar code
	_mm512_store_epi32(buf, rid);
	for (i = 0 ; i != 16 ; ++i) {
		size_t j = buf[i];
		if (j >= tuples) continue;
		const uint8_t *str = &strings[j * length];
		const uint8_t *str_len = &str[length];
		size_t s = ini_state;
		do {
			s <<= 8;
			s |= *str++;
			s = states[s];
		} while (0 <= (ssize_t) s && str != str_len);
		if (++s > rej_states) rids[o++] = j;
	}
	return o;
}

__attribute__((always_inline)) static inline
__m512i _mm512_broadcast_epi64(const int64_t *p)
{
	return _mm512_extload_epi64(p, _MM_UPCONV_EPI64_NONE,
	                               _MM_BROADCAST_1X8, 0);
}

size_t regex_vec_2(const uint8_t *strings,
                        uint32_t *rids,
                          size_t  tuples,
                          size_t  length,
                   const  int8_t *states,
                          size_t  ini_state,
                          size_t  rej_states)
{
	assert(tuples > 64);
	// constant vector masks
	const __mmask16 k0 = _mm512_int2mask(0);
	const __m512i m1s = _mm512_set1_epi32(-1);
	const __m512i m32 = _mm512_set1_epi32(32);
	const __m512i mFF = _mm512_set1_epi32(255);
	const __m512i len = _mm512_set1_epi32(length);
	const __m512i ini = _mm512_set1_epi32(ini_state);
	const __m512i rej = _mm512_set1_epi32(rej_states);
	const __m512i inc = _mm512_set_epi32(15,14,13,12,11,10, 9, 8,
	                                      7, 6, 5, 4, 3, 2, 1, 0);
	// aligned buffer
	uint32_t buf_unaligned[64 + 15];
	uint32_t *buf = (uint32_t*) ((63 + (size_t) buf_unaligned) & -64);
	// maintain out-of-order state
	const size_t loops_1 = length <= 4 ? length : 4;
	const size_t loops_2 = length <= 4 ? 0 :
	                      (length & 1) ? 1 :
	                      (length & 3) ? 2 : 4;
	size_t i = 16, o = 0, j = tuples - 17;
	__m512i rid_H = _mm512_add_epi32(inc, _mm512_set1_epi32(j));
	__m512i cur_L = ini, off_L = _mm512_set1_epi32(0);
	__m512i cur_H = ini, off_H = off_L, rid_L = inc;
	for (;;) {
		// compute (byte) offset using rid and string offset
		__m512i ptr_L = _mm512_fmadd_epi32(rid_L, len, off_L);
		__m512i ptr_H = _mm512_fmadd_epi32(rid_H, len, off_H);
		// gather 8 bytes per string
		__m512i ind_L = _mm512_srli_epi32(ptr_L, 2);
		__m512i ind_H = _mm512_srli_epi32(ptr_H, 2);
		__m512i word_1_L = _mm512_i32gather_epi32(ind_L, &strings[0], 4);
		__m512i word_2_L = _mm512_i32gather_epi32(ind_L, &strings[4], 4);
		__m512i word_1_H = _mm512_i32gather_epi32(ind_H, &strings[0], 4);
		__m512i word_2_H = _mm512_i32gather_epi32(ind_H, &strings[4], 4);
		if (length & 3) {
			// compute shifts
			__m512i shr_L = _mm512_slli_epi32(ptr_L, 30);
			__m512i shr_H = _mm512_slli_epi32(ptr_H, 30);
			shr_L = _mm512_srli_epi32(shr_L, 27);
			shr_H = _mm512_srli_epi32(shr_H, 27);
			__m512i shl_L = _mm512_sub_epi32(m32, shr_L);
			__m512i shl_H = _mm512_sub_epi32(m32, shr_H);
			// combine into words
			word_1_L = _mm512_srlv_epi32(word_1_L, shr_L);
			word_1_H = _mm512_srlv_epi32(word_1_H, shr_H);
			__m512i word_3_L = _mm512_sllv_epi32(word_2_L, shl_L);
			__m512i word_3_H = _mm512_sllv_epi32(word_2_H, shl_H);
			word_2_L = _mm512_srlv_epi32(word_2_L, shr_L);
			word_2_H = _mm512_srlv_epi32(word_2_H, shr_H);
			word_1_L = _mm512_or_epi32(word_1_L, word_3_L);
			word_1_H = _mm512_or_epi32(word_1_H, word_3_H);
		}
		// process 5 or 6 bytes per string
		size_t l = loops_1;
		__mmask16 k_L = _mm512_kxnor(k0, k0);
		__mmask16 k_H = _mm512_kxnor(k0, k0);
		do {
			// compute state transition index
			__m512i nex_L = _mm512_slli_epi32(cur_L, 8);
			__m512i nex_H = _mm512_slli_epi32(cur_H, 8);
			nex_L = _mm512_or_epi32(nex_L, _mm512_and_epi32(word_1_L, mFF));
			nex_H = _mm512_or_epi32(nex_H, _mm512_and_epi32(word_1_H, mFF));
			// gather new state for valid lanes
			cur_L = _mm512_mask_i32extgather_epi32(cur_L, k_L, nex_L, states,
			        _MM_UPCONV_EPI32_SINT8, 1, 0);
			cur_H = _mm512_mask_i32extgather_epi32(cur_H, k_H, nex_H, states,
			        _MM_UPCONV_EPI32_SINT8, 1, 0);
			// clear processed byte and update remaining bytes
			off_L = _mm512_mask_sub_epi32(off_L, k_L, off_L, m1s);
			off_H = _mm512_mask_sub_epi32(off_H, k_H, off_H, m1s);
			word_1_L = _mm512_srli_epi32(word_1_L, 8);
			word_1_H = _mm512_srli_epi32(word_1_H, 8);
			// update valid lanes
			k_L = _mm512_mask_cmpgt_epi32_mask(k_L, cur_L, m1s);
			k_H = _mm512_mask_cmpgt_epi32_mask(k_H, cur_H, m1s);
			k_L = _mm512_mask_cmpgt_epi32_mask(k_L, len, off_L);
			k_H = _mm512_mask_cmpgt_epi32_mask(k_H, len, off_H);
#ifdef _NO_LANE_VALID
			if (_mm512_kortestz(k_L, k_H)) goto no_lane_valid;
#endif
		} while (--l);
		if ((l = loops_2)) do {
			// compute state transition index
			__m512i nex_L = _mm512_slli_epi32(cur_L, 8);
			__m512i nex_H = _mm512_slli_epi32(cur_H, 8);
			nex_L = _mm512_or_epi32(nex_L, _mm512_and_epi32(word_2_L, mFF));
			nex_H = _mm512_or_epi32(nex_H, _mm512_and_epi32(word_2_H, mFF));
			// gather new state for valid lanes
			cur_L = _mm512_mask_i32extgather_epi32(cur_L, k_L, nex_L, states,
			        _MM_UPCONV_EPI32_SINT8, 1, 0);
			cur_H = _mm512_mask_i32extgather_epi32(cur_H, k_H, nex_H, states,
			        _MM_UPCONV_EPI32_SINT8, 1, 0);
			// clear processed byte and update remaining bytes
			off_L = _mm512_mask_sub_epi32(off_L, k_L, off_L, m1s);
			off_H = _mm512_mask_sub_epi32(off_H, k_H, off_H, m1s);
			word_2_L = _mm512_srli_epi32(word_2_L, 8);
			word_2_H = _mm512_srli_epi32(word_2_H, 8);
			// update valid lanes
			k_L = _mm512_mask_cmpgt_epi32_mask(k_L, cur_L, m1s);
			k_H = _mm512_mask_cmpgt_epi32_mask(k_H, cur_H, m1s);
			k_L = _mm512_mask_cmpgt_epi32_mask(k_L, len, off_L);
			k_H = _mm512_mask_cmpgt_epi32_mask(k_H, len, off_H);
#ifdef _NO_LANE_VALID
			if (_mm512_kortestz(k_L, k_H)) goto no_lane_valid;
#endif
		} while (--l);
#ifdef _NO_LANE_VALID
no_lane_valid:;
#endif
		// find processed and accepted strings
		__m512i out_L = _mm512_sub_epi32(cur_L, m1s);
		__m512i out_H = _mm512_sub_epi32(cur_H, m1s);
		__mmask16 ki_L = _mm512_knot(k_L);
		__mmask16 ki_H = _mm512_knot(k_H);
		__mmask16 ko_L = _mm512_mask_cmpgt_epu32_mask(ki_L, out_L, rej);
		__mmask16 ko_H = _mm512_mask_cmpgt_epu32_mask(ki_H, out_H, rej);
		// store rids of accepted strings
		if (!_mm512_kortestz(ko_L, ko_H)) {
			_mm512_mask_packstorelo_epi32(&rids[o +  0], ko_L, rid_L);
			_mm512_mask_packstorehi_epi32(&rids[o + 16], ko_L, rid_L);
			o += _mm_countbits_64(_mm512_kconcatlo_64(ko_L, k0));
			_mm512_mask_packstorelo_epi32(&rids[o +  0], ko_H, rid_H);
			_mm512_mask_packstorehi_epi32(&rids[o + 16], ko_H, rid_H);
			o += _mm_countbits_64(_mm512_kconcatlo_64(ko_H, k0));
		}
		// replace rids of processed strings
		if (!_mm512_kortestz(ki_L, ki_H)) {
			// update buffer with new rids
			buf[0] = i;
			j -= _mm_countbits_64(_mm512_kconcatlo_64(ki_H, k0));
			buf[1] = j;
			__m512i nex = _mm512_broadcast_epi64((int64_t*) buf);
			__m512i nex_L = _mm512_add_epi32(inc,
			                _mm512_swizzle_epi32(nex, _MM_SWIZ_REG_AAAA));
			__m512i nex_H = _mm512_add_epi32(inc,
			                _mm512_swizzle_epi32(nex, _MM_SWIZ_REG_BBBB));
			_mm512_store_epi32(&buf[ 0], nex_L);
			_mm512_store_epi32(&buf[16], nex_H);
			i += _mm_countbits_64(_mm512_kconcatlo_64(ki_L, k0));
			// replace finished rids
			rid_L = _mm512_mask_loadunpacklo_epi32(rid_L, ki_L, &buf[ 0]);
			rid_H = _mm512_mask_loadunpacklo_epi32(rid_H, ki_H, &buf[16]);
			off_L = _mm512_mask_xor_epi32(off_L, ki_L, off_L, off_L);
			off_H = _mm512_mask_xor_epi32(off_H, ki_H, off_H, off_H);
			cur_L = _mm512_mask_blend_epi32(ki_L, cur_L, ini);
			cur_H = _mm512_mask_blend_epi32(ki_H, cur_H, ini);
			// stop if end of tuples is reached
			if (j < i + 32) break;
		}
	}
	// process last strings with scalar code
	_mm512_store_epi32(&buf[ 0], rid_L);
	_mm512_store_epi32(&buf[16], rid_H);
	buf[32] = tuples - 1;
	size_t l = 33;
	assert(i <= j);
	while (i != j) buf[l++] = i++;
	assert(l <= 64);
	// process last rids in scalar code
	for (i = 0 ; i != l ; ++i) {
		j = buf[i];
		const uint8_t *str = &strings[j * length];
		const uint8_t *str_len = &str[length];
		size_t s = ini_state;
		do {
			s <<= 8;
			s |= *str++;
			s = states[s];
		} while (0 <= (ssize_t) s && str != str_len);
		if (++s > rej_states) rids[o++] = j;
	}
	return o;
}

size_t regex_vec_3(const uint8_t *strings,
                        uint32_t *rids,
                          size_t  tuples,
                          size_t  length,
                   const  int8_t *states,
                          size_t  ini_state,
                          size_t  rej_states)
{
	return 0;
}

#endif

uint64_t thread_time(void)
{
	struct timespec t;
	assert(clock_gettime(CLOCK_THREAD_CPUTIME_ID, &t) == 0);
	return t.tv_sec * 1000 * 1000 * 1000 + t.tv_nsec;
}

int uint32_cmp(const void *p1, const void *p2)
{
	uint32_t i1 = * (const uint32_t*) p1;
	uint32_t i2 = * (const uint32_t*) p2;
	return i1 < i2 ? -1 : i1 > i2 ? 1 : 0;
}

typedef struct {
	pthread_t id;
	uint32_t seed;
	size_t length;
	size_t tuples;
	size_t loops;
	size_t tuples_out;
	size_t ini_state;
	size_t rej_states;
	const int8_t *states;
	const char *charset;
	uint64_t *time_ns;
	pthread_barrier_t *barrier;
} thread_info_t;

void *run(void *arg)
{
	thread_info_t *info = (thread_info_t*) arg;
	assert(pthread_equal(pthread_self(), info->id));
	pthread_barrier_t *barrier = info->barrier;
	// copy info
	uint32_t seed = info->seed;
	size_t tuples = info->tuples;
	size_t length = info->length;
	const char *charset = info->charset;
	const int8_t *states = info->states;
	size_t ini_state  = info->ini_state;
	size_t rej_states = info->rej_states;
	// generate input
	size_t i, l = tuples * length;
	assert((l >> 32) == 0);
	uint8_t *strings = malloc(l + 4);
	assert(strings != NULL);
	for (i = 0 ; i < l ; i += 5) {
		size_t j = rand_r(&seed);
		strings[i + 0] = charset[(j >> 24) & 63];
		strings[i + 1] = charset[(j >> 18) & 63];
		strings[i + 2] = charset[(j >> 12) & 63];
		strings[i + 3] = charset[(j >>  6) & 63];
		strings[i + 4] = charset[(j >>  0) & 63];
	}
	// run scalar to count output
	i = regex_cnt(strings, tuples, length, &l,
	              states, ini_state, rej_states);
	info->loops = l;
	info->tuples_out = i;
	// output space
	i += 16;
	uint32_t *rids_sca = malloc(i << 2);
	uint32_t *rids_vec = malloc(i << 2);
	while (i--) {
		rids_sca[i] = 0xCAFEBABE;
		rids_vec[i] = 0xDEADBEEF;
	}
	// run scalar
	pthread_barrier_wait(barrier++);
	uint64_t t = thread_time();
	i = regex_sca(strings, rids_sca, tuples, length,
	              states, ini_state, rej_states);
	t = thread_time() - t;
	assert(i == info->tuples_out);
	info->time_ns[0] = t;
	// run vector 1
	pthread_barrier_wait(barrier++);
	t = thread_time();
	i = regex_vec_1(strings, rids_vec, tuples, length,
	                states, ini_state, rej_states);
	t = thread_time() - t;
	assert(i == info->tuples_out);
	info->time_ns[1] = t;
	// verify rids of vector 1
	pthread_barrier_wait(barrier++);
	qsort(rids_vec, info->tuples_out, 4, uint32_cmp);
	for (i = 0 ; i != info->tuples_out ; ++i)
		assert(rids_sca[i] == rids_vec[i]);
	// run vector 2
	pthread_barrier_wait(barrier++);
	t = thread_time();
	i = regex_vec_2(strings, rids_vec, tuples, length,
	                states, ini_state, rej_states);
	t = thread_time() - t;
	assert(i == info->tuples_out);
	info->time_ns[2] = t;
	// verify rids of vector 2
	pthread_barrier_wait(barrier++);
	qsort(rids_vec, info->tuples_out, 4, uint32_cmp);
	for (i = 0 ; i != info->tuples_out ; ++i)
		assert(rids_sca[i] == rids_vec[i]);
	// run vector 3
	pthread_barrier_wait(barrier++);
	t = thread_time();
	i = regex_vec_3(strings, rids_vec, tuples, length,
	                states, ini_state, rej_states);
	t = thread_time() - t;
	assert(i == info->tuples_out);
	info->time_ns[3] = t;
	// verify rids of vector 3
	pthread_barrier_wait(barrier++);
	qsort(rids_vec, info->tuples_out, 4, uint32_cmp);
	for (i = 0 ; i != info->tuples_out ; ++i)
		assert(rids_sca[i] == rids_vec[i]);
	// cleanup
	pthread_barrier_wait(barrier++);
	free(strings);
	free(rids_sca);
	free(rids_vec);
	pthread_exit(NULL);
}

void reject_accept_now(const int8_t *states, size_t nodes,
                       size_t *rej, size_t *acc)
{
	size_t a, r, i, e = nodes << 8;
	for (a = r = i = 0 ; i != e ; ++i) {
		int8_t s = states[i];
		r += (s == -1) ? 1 : 0;
		a += (s <  -1) ? 1 : 0;
	}
	*rej = r;
	*acc = a;
}

int8_t *dfa_compress(const int32_t *states_int, size_t nodes)
{
	if (nodes >= 128) return NULL;
	size_t i, j = nodes << 8;
	int8_t *states_byte = malloc(j);
	for (i = 0 ; i != j ; ++i) {
		int32_t si = states_int[i];
		int8_t sb = (int8_t) si;
		if (si != (int32_t) sb) break;
		states_byte[i] = sb;
	}
	if (i != j) free(states_byte);
	return i != j ? NULL : states_byte;
}

int hardware_threads(void)
{
	char name[48];
	struct stat st;
	int threads = -1;
	do {
		sprintf(name, "/sys/devices/system/cpu/cpu%d", ++threads);
	} while (stat(name, &st) == 0);
	assert(threads > 0);
	return threads;
}

int main(int argc, char **argv)
{
	// default charset
	const char *def_charset = "@.1234567890"
	                          "abcdefghijklmnopqrstuvwxyz"
	                          "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
	// process arguments
	double size_GB = argc > 1 ? atof(argv[1]) : 10.0;
	size_t length = argc > 2 ? atoll(argv[2]) : 30;
	int t = hardware_threads();
	int threads = argc > 3 ? atoi(argv[3]) : t;
	const char *charset = argc > 4 ? argv[4] : def_charset;
	assert(strlen(charset) == 64);
	// get DFA
	size_t ini_state, rej_states, acc_states;
	const int32_t *states_int = dfa(&ini_state, &rej_states, &acc_states);
	if (states_int == NULL) {
		fprintf(stderr, "Invalid DFA\n");
		return EXIT_FAILURE;
	}
	size_t all_states = acc_states + rej_states;
	if (all_states >= 128) {
		fprintf(stderr, "DFA has too many states: %ld\n", all_states);
		return EXIT_SUCCESS;
	}
	// compress DFA
	const int8_t *states = dfa_compress(states_int, all_states);
	free((void*) states_int);
	assert(states != NULL);
	// count accept now and reject now transitions
	size_t rej_now, acc_now;
	reject_accept_now(states, all_states, &rej_now, &acc_now);
	// compute number of tuples
	size_t tuples = (size_GB * 1024 * 1024 * 1024) / length;
	tuples -= tuples % (threads << 4);
	// print info
#ifdef __MIC
	double bandwidth = 212;
#else
	double bandwidth = 21.8;
#endif
	fprintf(stderr, "Threads: %d / %d\n", threads, t);
	fprintf(stderr, "Tuples: %ld\n", tuples);
	fprintf(stderr, "Charset: %s\n", charset);
	fprintf(stderr, "Reject states: %ld\n", rej_states + (rej_now ? 1 : 0));
	fprintf(stderr, "Accept states: %ld\n", acc_states + (acc_now ? 1 : 0));
	fprintf(stderr, "DFA footprint: %.1f KB\n", (all_states << 8) / 1024.0);
	fprintf(stderr, "RAM bandwidth: %.1f GB/s\n", bandwidth);
#ifdef _NO_LANE_VALID
	fprintf(stderr, "No valid lane branch: enabled\n");
#else
	fprintf(stderr, "No valid lane branch: disabled\n");
#endif
	fprintf(stderr, "Input size: %.3f GB\n", size_GB);
	fprintf(stderr, "Length (max): %ld\n", length);
	// run threads
	const char *method[] = {"Scalar", "Vector (x1)", "Vector (x2)", "Vector (ls)"};
	int m, methods = sizeof(method) / sizeof(char*);
	int b, barriers = methods * 2;
	pthread_barrier_t barrier[barriers];
	for (b = 0 ; b != barriers ; ++b)
		pthread_barrier_init(&barrier[b], NULL, threads);
	srand(time(NULL));
	uint64_t time_ns[methods * threads];
	thread_info_t info[threads];
	for (t = 0 ; t != threads ; ++t) {
		info[t].seed = rand();
		info[t].tuples = tuples / threads;
		info[t].length = length;
		info[t].states = states;
		info[t].charset = charset;
		info[t].ini_state  = ini_state;
		info[t].rej_states = rej_states;
		info[t].time_ns = &time_ns[methods * t];
		info[t].barrier = barrier;
		pthread_create(&info[t].id, NULL, run, &info[t]);
	}
	size_t all_loops = 0;
	size_t tuples_out = 0;
	for (t = 0 ; t != threads ; ++t) {
		pthread_join(info[t].id, NULL);
		tuples_out += info[t].tuples_out;
		all_loops += info[t].loops;
	}
	free((void*) states);
	for (b = 0 ; b != barriers ; ++b)
		pthread_barrier_destroy(&barrier[b]);
	// print results
	size_t rej_loops = all_loops - tuples_out * length;
	fprintf(stderr, "Length (all): %.2f\n", all_loops * 1.0 / tuples);
	fprintf(stderr, "Length (rej): %.2f\n", rej_loops * 1.0 / tuples);
	fprintf(stderr, "Selectivity: %.4f%%\n", tuples_out * 100.0 / tuples);
	double s_0 = 0.0;
	for (m = 0 ; m != methods ; ++m) {
		uint64_t ns = 0;
		for (t = 0 ; t != threads ; ++t)
			ns += info[t].time_ns[m];
		double s = (ns / 1.0e9) / threads;
		double r = size_GB / s;
		fprintf(stderr, "Method: %s\n", method[m]);
		fprintf(stderr, "Time: %.3f sec\n", s);
		fprintf(stderr, "Rate: %.3f GB/s (%.1f%%)\n", r, r * 100.0 / bandwidth);
		if (m) fprintf(stderr, "Speedup: %.2fX\n", s_0 / s);
		else s_0 = s;
	}
	return EXIT_SUCCESS;
}
