/* Copyright (c) 2016
 * The Trustees of Columbia University in the City of New York
 * All rights reserved.
 *
 * Authors:  Orestis Polychroniou  (orestis@cs.columbia.edu)
 *           Eva Sitaridi (eva@cs.columbia.edu)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <time.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <stdint.h>
#include <pthread.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <smmintrin.h>


inline uint8_t* __attribute__((always_inline))
sse(const uint8_t *str,
          __m128i  pat,
           size_t  str_len,
           size_t  pat_len)
{
	size_t i = 0, l = 16 - pat_len;
	while (i + 16 < str_len) {
		__m128i s = _mm_loadu_si128((__m128i*) &str[i]);
		size_t j = _mm_cmpistri(pat, s, 12);
		if (j >= 16) i += 16;
		else {
			if (j <= l)
				return (uint8_t*) &str[i + pat_len];
			i += j;
		}
	}
	if (i + pat_len <= str_len) {
		__m128i s = _mm_loadu_si128((__m128i*) &str[i]);
		size_t j = _mm_cmpestri(pat, pat_len, s, str_len - i, 12);
		if (j < 16 && j <= l)
			return (uint8_t*) &str[i + pat_len];
	}
	return NULL;
}

inline uint8_t* __attribute__((always_inline))
kmp(const uint8_t *str,
    const uint8_t *pat,
           size_t  str_len,
           size_t  pat_len,
     const size_t *jmp)
{
	size_t i = 0, j = 0;
	while (str_len + j >= pat_len + i) {
		uint8_t c = str[i++];
		if (c != pat[j]) {
			while (j) {
				j = jmp[j];
				if (c == pat[j]) {
					j++;
					break;
				}
			}
		} else if (++j == pat_len)
			return (uint8_t*) &str[i];
	}
	return NULL;
}

inline size_t __attribute__((always_inline))
max(size_t x, size_t y)
{
	asm("cmpq	%1, %0\n\t"
	    "cmovcq	%1, %0"
	: "=r"(x) : "r"(y), "0"(x) : "cc");
	return x;
}

inline uint8_t* __attribute__((always_inline))
bm(const uint8_t *str,
   const uint8_t *pat,
          size_t  str_len,
          size_t  pat_len,
   const  size_t *sym_jmp,
   const  size_t *pat_jmp)
{
	size_t i = pat_len - 1;
	while (i < str_len) {
		size_t j = pat_len - 1;
		uint8_t b = str[i];
		while (b == pat[j]) {
			if (j == 0) return (uint8_t*) &str[i + pat_len];
			b = str[--i], j--;
		}
		i += max(pat_jmp[j], sym_jmp[b]);
	}
	return NULL;
}

inline uint8_t* __attribute__((always_inline))
dfa(const uint8_t *str,
    const uint8_t *pat,
           size_t  str_len,
           size_t  pat_len,
    const uint8_t *dfa_jmp)
{
	size_t i = 0, j = 0;
	while (str_len + j >= pat_len + i) {
		j <<= 8;
		j |= str[i++];
		j = dfa_jmp[j];
		if (j == pat_len)
			return (uint8_t*) &str[i];
	}
	return NULL;
}

size_t *kmp_jmp(const uint8_t *pat, size_t pat_len)
{
	size_t i = 0, j = 2;
	size_t *jmp = malloc(max(2, pat_len) * sizeof(size_t));
	jmp[0] = -1;
	jmp[1] = 0;
	while (j < pat_len)
		if (pat[j - 1] == pat[i])
			jmp[j++] = ++i;
		else if (i != 0 && i != -1)
			i = jmp[i];
		else
			jmp[j++] = 0;
	return jmp;
}

uint8_t *kmp_dfa_jmp(const uint8_t *pat, size_t pat_len, const size_t *jmp)
{
	size_t i, j;
	assert(pat_len < 128);
	uint8_t *s = malloc(pat_len << 8);
	memset(s, 0, pat_len << 8);
	for (i = 0 ; i != pat_len ; ++i)
		s[(i << 8) | pat[i]] = i + 1;
	for (i = 1 ; i != pat_len ; ++i) {
		j = jmp[i];
		do {
			uint8_t c = (uint8_t) pat[j];
			s[(i << 8) | c] = j + 1;
			j = jmp[j];
		} while (j != -1);
	}
	return s;
}

size_t *bm_sym_jmp(const uint8_t *pat, size_t pat_len)
{
	size_t i, *sym_jmp = malloc(256 * sizeof(size_t));
	for (i = 0 ; i != 256 ; ++i)
		sym_jmp[i] = pat_len;
	for (i = 0 ; i != pat_len - 1 ; ++i)
		sym_jmp[pat[i]] = pat_len - 1 - i;
	return sym_jmp;
}

int bm_is_prefix(const uint8_t *word, size_t word_len, size_t p)
{
	size_t i;
	for (i = 0 ; i != word_len - p ; ++i)
		if (word[i] == word[i + p]) return 0;
	return 1;
}

size_t bm_suffix_length(const uint8_t *word, size_t word_len, size_t p)
{
	size_t i = 0;
	while (word[p - i] == word[word_len - 1 - i] && i < p) i++;
	return i;
}

size_t *bm_pat_jmp(const uint8_t *pat, size_t pat_len)
{
	size_t p, l = pat_len - 1;
	size_t *pat_jmp = malloc(pat_len * sizeof(size_t));
	for (p = l ; p != -1 ; --p) {
		if (bm_is_prefix(pat, pat_len, p + 1)) l = p + 1;
		pat_jmp[p] = l + pat_len - 1 - p;
	}
	for (p = 0 ; p != pat_len - 1 ; ++p) {
		size_t s = bm_suffix_length(pat, pat_len, p);
		if (pat[p - s] != pat[pat_len - 1 - s])
			pat_jmp[pat_len - 1 - s] = s + pat_len - 1 - p;
	}
	return pat_jmp;
}

uint64_t thread_time(void)
{
	struct timespec t;
	assert(clock_gettime(CLOCK_THREAD_CPUTIME_ID, &t) == 0);
	return t.tv_sec * 1000 * 1000 * 1000 + t.tv_nsec;
}

typedef struct {
	pthread_t id;
	int thread;
	int threads;
	const uint8_t *str;
	const uint8_t *str_len;
	size_t pat_1_len;
	size_t pat_2_len;
	const uint8_t *pat_1;
	const uint8_t *pat_2;
	const size_t *kmp_jmp_1;
	const size_t *kmp_jmp_2;
	const size_t *bm_sym_jmp_1;
	const size_t *bm_sym_jmp_2;
	const size_t *bm_pat_jmp_1;
	const size_t *bm_pat_jmp_2;
	const uint8_t *dfa_jmp_1;
	const uint8_t *dfa_jmp_2;
	uint64_t *bitmap_1;
	uint64_t *bitmap_2;
	size_t bytes;
	size_t tuples;
	size_t *bytes_per_thread;
	size_t *tuples_per_thread;
	uint64_t *times_ns;
	pthread_barrier_t *barrier;
} thread_info_t;

void *run(void *arg)
{
	thread_info_t *info = (thread_info_t*) arg;
	assert(pthread_equal(pthread_self(), info->id));
	pthread_barrier_t *barrier = info->barrier;
	size_t i, j;
	// copy basic data
	size_t thread  = info->thread;
	size_t threads = info->threads;
	size_t tuples  = info->tuples;
	const uint8_t *str     = info->str;
	const uint8_t *str_len = info->str_len;
	const uint8_t *pat_1 = info->pat_1;
	const uint8_t *pat_2 = info->pat_2;
	uint64_t *bitmap_1 = info->bitmap_1;
	uint64_t *bitmap_2 = info->bitmap_2;
	size_t pat_1_len = info->pat_1_len;
	size_t pat_2_len = info->pat_2_len;
	const size_t *kmp_jmp_1 = info->kmp_jmp_1;
	const size_t *kmp_jmp_2 = info->kmp_jmp_2;
	const size_t *bm_sym_jmp_1 = info->bm_sym_jmp_1;
	const size_t *bm_sym_jmp_2 = info->bm_sym_jmp_2;
	const size_t *bm_pat_jmp_1 = info->bm_pat_jmp_1;
	const size_t *bm_pat_jmp_2 = info->bm_pat_jmp_2;
	const uint8_t *dfa_jmp_1 = info->dfa_jmp_1;
	const uint8_t *dfa_jmp_2 = info->dfa_jmp_2;
	// local thread strings
	size_t tuples_per_thread = (tuples / threads) & -64;
	size_t thread_beg = tuples_per_thread * (thread + 0);
	size_t thread_end = tuples_per_thread * (thread + 1);
	if (thread + 1 == threads) thread_end = tuples;
	// compute length sum of strings
	for (i = thread_beg, j = 0 ; i != thread_end ; ++i)
		j += str_len[i];
	info->bytes_per_thread[thread] = j;
	// sync
	pthread_barrier_wait(barrier++);
	// find starting offset per string
	for (i = j = 0 ; i != thread ; ++i)
		j += info->bytes_per_thread[i];
	const uint8_t *str_beg = str + j;
	j += info->bytes_per_thread[i++];
	const uint8_t *str_end = str + j;
	for (; i != threads ; ++i)
		j += info->bytes_per_thread[i];
	assert(j == info->bytes);
	// sync
	pthread_barrier_wait(barrier++);
	// KMP scalar code
	uint64_t t = thread_time();
	const uint8_t *s = str_beg;
	uint64_t *b = &bitmap_1[thread_beg >> 6];
	for (i = thread_beg ; i < thread_end ; i += 64) {
		uint64_t m = 0;
		size_t n = thread_end - i;
		if (n > 64) n = 64;
		for (j = 0 ; j != n ; ++j) {
			size_t s_l = str_len[i + j];
			const uint8_t *ss = kmp(s, pat_1, s_l, pat_1_len, kmp_jmp_1);
			if (ss != NULL) {
				size_t ss_l = s_l - (ss - s);
				if (kmp(ss, pat_2, ss_l, pat_2_len, kmp_jmp_2) != NULL)
					m |= ((uint64_t) 1) << j;
			}
			s += s_l;
		}
		_mm_stream_si64((long long*) b++, m);
	}
	assert(s == str_end);
	t = thread_time() - t;
	info->times_ns[0] = t;
	// sync
	pthread_barrier_wait(barrier++);
	// count set bits
	uint64_t c = 0;
	for (i = thread_beg ; i < thread_end ; i += 64)
		c += _mm_popcnt_u64(bitmap_1[i >> 6]);
	info->tuples_per_thread[thread] = (size_t) c;
	// sync
	pthread_barrier_wait(barrier++);
	// DFA scalar code
	t = thread_time();
	s = str_beg;
	b = &bitmap_2[thread_beg >> 6];
	for (i = thread_beg ; i < thread_end ; i += 64) {
		uint64_t m = 0;
		size_t n = thread_end - i;
		if (n > 64) n = 64;
		for (j = 0 ; j != n ; ++j) {
			size_t s_l = str_len[i + j];
			const uint8_t *ss = dfa(s, pat_1, s_l, pat_1_len, dfa_jmp_1);
			if (ss != NULL) {
				size_t ss_l = s_l - (ss - s);
				if (dfa(ss, pat_2, ss_l, pat_2_len, dfa_jmp_2) != NULL)
					m |= ((uint64_t) 1) << j;
			}
			s += s_l;
		}
		_mm_stream_si64((long long*) b++, m);
	}
	assert(s == str_end);
	t = thread_time() - t;
	info->times_ns[1] = t;
	// sync
	pthread_barrier_wait(barrier++);
	// verify that bitmaps are equal
	for (i = thread_beg ; i < thread_end ; i += 64)
		assert(bitmap_1[i >> 6] == bitmap_2[i >> 6]);
	// sync
	pthread_barrier_wait(barrier++);
	// BM scalar code
	t = thread_time();
	s = str_beg;
	b = &bitmap_2[thread_beg >> 6];
	for (i = thread_beg ; i < thread_end ; i += 64) {
		uint64_t m = 0;
		size_t n = thread_end - i;
		if (n > 64) n = 64;
		for (j = 0 ; j != n ; ++j) {
			size_t s_l = str_len[i + j];
			const uint8_t *ss = bm(s, pat_1, s_l, pat_1_len,
			                       bm_sym_jmp_1, bm_pat_jmp_1);
			if (ss != NULL) {
				size_t ss_l = s_l - (ss - s);
				if (bm(ss, pat_2, ss_l, pat_2_len,
				       bm_sym_jmp_2, bm_pat_jmp_2) != NULL)
					m |= ((uint64_t) 1) << j;
			}
			s += s_l;
		}
		_mm_stream_si64((long long*) b++, m);
	}
	assert(s == str_end);
	t = thread_time() - t;
	info->times_ns[2] = t;
	// sync
	pthread_barrier_wait(barrier++);
	// verify that bitmaps are equal
	for (i = thread_beg ; i < thread_end ; i += 64)
		assert(bitmap_1[i >> 6] == bitmap_2[i >> 6]);
	// sync
	pthread_barrier_wait(barrier++);
	// SSE vector code
	t = thread_time();
	s = str_beg;
	b = &bitmap_2[thread_beg >> 6];
	__m128i p1 = _mm_loadu_si128((__m128i*) pat_1);
	__m128i p2 = _mm_loadu_si128((__m128i*) pat_2);
	for (i = thread_beg ; i < thread_end ; i += 64) {
		uint64_t m = 0;
		size_t n = thread_end - i;
		if (n > 64) n = 64;
		for (j = 0 ; j != n ; ++j) {
			size_t s_l = str_len[i + j];
			const uint8_t *ss = sse(s, p1, s_l, pat_1_len);
			if (ss != NULL) {
				size_t ss_l = s_l - (ss - s);
				if (sse(ss, p2, ss_l, pat_2_len) != NULL)
					m |= ((uint64_t) 1) << j;
			}
			s += s_l;
		}
		_mm_stream_si64((long long*) b++, m);
	}
	assert(s == str_end);
	t = thread_time() - t;
	info->times_ns[3] = t;
	// sync
	pthread_barrier_wait(barrier++);
	// verify that bitmaps are equal
	for (i = thread_beg ; i < thread_end ; i += 64)
		assert(bitmap_1[i >> 6] == bitmap_2[i >> 6]);
	pthread_exit(NULL);
}

int hardware_threads(void)
{
	char name[48];
	struct stat st;
	int threads = -1;
	do {
		sprintf(name, "/sys/devices/system/cpu/cpu%d", ++threads);
	} while (stat(name, &st) == 0);
	assert(threads > 0);
	return threads;
}

int parse(const char *file, size_t column,
          uint8_t *arr[2], size_t arr_size[2])
{
	FILE *fp = fopen(file, "rb");
	if (fp == NULL) return 0;
	size_t str_i = 0, str_c = 128;
	size_t len_i = 0, len_c = 16;
	uint8_t *str = malloc(str_c);
	uint8_t *len = malloc(len_c);
	const int column_delim = '|';
	const int tuple_delim = '\n';
	int ch;
	while ((ch = fgetc(fp)) != EOF) {
		// skip columns
		size_t col = column;
		while (col) {
			while (ch != column_delim) {
				if (ch == tuple_delim || ch == EOF) goto error;
				ch = fgetc(fp);
			}
			ch = fgetc(fp);
			col--;
		}
		// copy actual string data
		size_t str_i_old = str_i;
		while (ch != column_delim) {
			if (ch == tuple_delim || ch == EOF) goto error;
			str[str_i++] = ch;
			if (str_i == str_c) {
				str_c += str_c;
				str = realloc(str, str_c);
			}
			ch = fgetc(fp);
		}
		// append string length
		len[len_i++] = str_i - str_i_old;
		if (len_i == len_c) {
			len_c += len_c;
			len = realloc(len, len_c);
		}
		// discard the rest of the tuple
		while ((ch = fgetc(fp)) != tuple_delim)
			if (ch == EOF) goto error;
	}
	fclose(fp);
	arr[0] = realloc(str, str_i);
	arr[1] = realloc(len, len_i);
	arr_size[0] = str_i;
	arr_size[1] = len_i;
	return 1;
error:
	fclose(fp);
	free(str);
	free(len);
	return 0;
}

uint8_t *byte_copy(const char *src, size_t *len)
{
	size_t i = 0;
	while (src[i]) i++;
	*len = i;
	uint8_t *dst = malloc(max(i, 16));
	while (i--) dst[i] = src[i];
	return dst;
}

int main(int argc, char **argv)
{
	// check arguments
	if (argc < 5 || argc > 8) {
		fprintf(stderr, "Usage: %s file column pattern_1 pattern_2 "
		                "[scale_factor] [threads] [repeats]\n", argv[0]);
		return EXIT_FAILURE;
	}
	int max_threads = hardware_threads();
	const char *file = argv[1];
	size_t column = atoll(argv[2]);
	const char *pattern_1 = argv[3];
	const char *pattern_2 = argv[4];
	size_t s, scale_factor = argc > 5 ? atoll(argv[5]) : 1;
	int t, threads = argc > 6 ? atoi(argv[6]) : max_threads;
	size_t r, repeats = argc > 7 ? atoll(argv[7]) : 1;
	assert(scale_factor >= 1);
	assert(threads >= 1 && threads <= max_threads);
	// copy patterns to byte arrays and check length
	size_t pat_1_len, pat_2_len;
	uint8_t *pat_1 = byte_copy(pattern_1, &pat_1_len);
	uint8_t *pat_2 = byte_copy(pattern_2, &pat_2_len);
	assert(pat_1_len >= 2 && pat_1_len <= 16);
	assert(pat_2_len >= 2 && pat_2_len <= 16);
	// parse database file and extract target column
	uint8_t *arr[2];
	size_t arr_size[2];
	if (!parse(file, column, arr, arr_size)) {
		fprintf(stderr, "%s: %s: Parse error\n", argv[0], file);
		return EXIT_FAILURE;
	}
	uint8_t *str = arr[0];
	uint8_t *str_len = arr[1];
	size_t bytes = arr_size[0];
	size_t tuples = arr_size[1];
	// resize the arrays using scale factor
	if (scale_factor != 1) {
		str     = realloc(str,     scale_factor * bytes);
		str_len = realloc(str_len, scale_factor * tuples);
	}
	for (s = 1 ; s != scale_factor ; ++s) {
		memcpy(&str    [s * bytes],  str,     bytes);
		memcpy(&str_len[s * tuples], str_len, tuples);
	}
	bytes  *= scale_factor;
	tuples *= scale_factor;
	// print info
	double gb = (tuples + bytes) * 1.0 / (1 << 30);
	fprintf(stderr, "Threads: %d / %d\n", threads, max_threads);
	fprintf(stderr, "Tuples: %ld\n", tuples);
	fprintf(stderr, "Bytes: %ld\n", bytes);
	fprintf(stderr, "Length: %.3f\n", bytes * 1.0 / tuples);
	fprintf(stderr, "Space: %.2f GB (SF = %ld)\n", gb, scale_factor);
	fprintf(stderr, "Pattern 1: \"%s\"\n", pattern_1);
	fprintf(stderr, "Pattern 2: \"%s\"\n", pattern_2);
	// compute jump arrays for KMP
	size_t *kmp_jmp_1 = kmp_jmp(pat_1, pat_1_len);
	size_t *kmp_jmp_2 = kmp_jmp(pat_2, pat_2_len);
	// compute jump arrays for BM
	size_t *bm_sym_jmp_1 = bm_sym_jmp(pat_1, pat_1_len);
	size_t *bm_sym_jmp_2 = bm_sym_jmp(pat_2, pat_2_len);
	size_t *bm_pat_jmp_1 = bm_pat_jmp(pat_1, pat_1_len);
	size_t *bm_pat_jmp_2 = bm_pat_jmp(pat_2, pat_2_len);
	// compute DFAs
	uint8_t *dfa_jmp_1 = kmp_dfa_jmp(pat_1, pat_1_len, kmp_jmp_1);
	uint8_t *dfa_jmp_2 = kmp_dfa_jmp(pat_2, pat_2_len, kmp_jmp_2);
	// allocate arrays for bitmaps
	size_t w, bitmap_words = (tuples + 63) >> 6;
	uint64_t *bitmap_1 = malloc(bitmap_words << 3);
	uint64_t *bitmap_2 = malloc(bitmap_words << 3);
	for (w = 0 ; w != bitmap_words ; ++w) {
		bitmap_1[w] = 0xABCD1234ABCD1234ull;
		bitmap_2[w] = 0xDBCA8765DBCA8765ull;
	}
	// repeat the experiment
	for (r = 0 ; r != repeats ; ++r) {
		// methods
		const char *method[] = {"KMP", "DFA", "BM", "SSE"};
		int m, methods = sizeof(method) / sizeof(char*);
		// run threads
		int b, barriers = methods * 2 + 1;
		pthread_barrier_t barrier[barriers];
		for (b = 0 ; b != barriers ; ++b)
			pthread_barrier_init(&barrier[b], NULL, threads);
		size_t  bytes_per_thread[threads];
		size_t tuples_per_thread[threads];
		uint64_t times_ns[threads * methods];
		thread_info_t info[threads];
		for (t = 0 ; t != threads ; ++t) {
			info[t].thread = t;
			info[t].threads = threads;
			info[t].tuples = tuples;
			info[t].bytes = bytes;
			info[t].pat_1 = pat_1;
			info[t].pat_2 = pat_2;
			info[t].pat_1_len = pat_1_len;
			info[t].pat_2_len = pat_2_len;
			info[t].kmp_jmp_1 = kmp_jmp_1;
			info[t].kmp_jmp_2 = kmp_jmp_2;
			info[t].bm_sym_jmp_1 = bm_sym_jmp_1;
			info[t].bm_sym_jmp_2 = bm_sym_jmp_2;
			info[t].bm_pat_jmp_1 = bm_pat_jmp_1;
			info[t].bm_pat_jmp_2 = bm_pat_jmp_2;
			info[t].dfa_jmp_1 = dfa_jmp_1;
			info[t].dfa_jmp_2 = dfa_jmp_2;
			info[t].str = str;
			info[t].str_len = str_len;
			info[t].bitmap_1 = bitmap_1;
			info[t].bitmap_2 = bitmap_2;
			info[t].bytes_per_thread = bytes_per_thread;
			info[t].tuples_per_thread = tuples_per_thread;
			info[t].barrier = barrier;
			info[t].times_ns = &times_ns[methods * t];
			pthread_create(&info[t].id, NULL, run, (void*) &info[t]);
		}
		size_t tuples_ss = 0;
		for (t = 0 ; t != threads ; ++t) {
			pthread_join(info[t].id, NULL);
			tuples_ss += info[t].tuples_per_thread[t];
		}
		// print info
		fprintf(stderr, "\nRepeat: %ld / %ld\n", r + 1, repeats);
		fprintf(stderr, "Selectivity: %.5f%%\n",
		                 tuples_ss * 100.0 / tuples);
		for (m = 0 ; m != methods ; ++m) {
			uint64_t ns = 0;
			for (t = 0 ; t != threads ; ++t)
				ns += info[t].times_ns[m];
			double s = (ns / 1.0e9) / threads;
			fprintf(stderr, "%-3s: %.3f s | %.2f GB/s\n",
			                 method[m], s, gb / s);
		}
		// cleanup
		for (b = 0 ; b != barriers ; ++b)
			pthread_barrier_destroy(&barrier[b]);
	}
	free(str);
	free(str_len);
	free(bitmap_1);
	free(bitmap_2);
	free(pat_1);
	free(pat_2);
	free(kmp_jmp_1);
	free(kmp_jmp_2);
	free(dfa_jmp_1);
	free(dfa_jmp_2);
	free(bm_sym_jmp_1);
	free(bm_sym_jmp_2);
	free(bm_pat_jmp_1);
	free(bm_pat_jmp_2);
	return EXIT_SUCCESS;
}
